"""In-process tests for the flatpak generator"""

from contextlib import contextmanager
import logging
import os
import tempfile
from unittest.mock import Mock, patch

from click.testing import CliRunner
import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd
import pytest
import responses
import yaml

from _fedmod.cli import _cli_commands


log = logging.getLogger(__name__)

testfiles_dir = os.path.join(os.path.dirname(__file__), 'files', 'flatpak')

with open(os.path.join(testfiles_dir, 'apps.json')) as f:
    APPS_JSON = f.read()

with open(os.path.join(testfiles_dir, 'eog.yaml')) as f:
    EOG_YAML = f.read()

with open(os.path.join(testfiles_dir, 'eog.json')) as f:
    EOG_JSON = f.read()

with open(os.path.join(testfiles_dir, 'flatpak-runtime.mmd.yaml')) as f:
    FLATPAK_RUNTIME_MODULEMD = f.read()


@contextmanager
def mock_koji():
    session = Mock()
    session.getPackageID = Mock(return_value=42)
    session.listBuilds = Mock(return_value=[{
        'name': 'flatpak-runtime',
        'version': 'f29',
        'release': '20180831153244.1',
        'build_id': '1167601',
        'extra': {
            'typeinfo': {
                'module': {
                    'modulemd_str': FLATPAK_RUNTIME_MODULEMD
                }
            }
        }
    }])
    session.listTags = Mock(return_value=[
        {'name': 'f29-modular-updates-candidate'}
    ])
    # listArchives/listRPMs are set up to return the RPMs for a mocked
    # flatpak-common module - the generator code *either* looks up
    # the modulemd or it looks up the RPM list, not both, so we don't
    # have to pay attention to arguments passed.
    session.listArchives = Mock(return_value=[
        {'btype': 'module',
         'build_id': 1149156,
         'filename': 'modulemd.txt',
         'id': 160909}
    ])
    session.listRPMs = Mock(return_value=[
        {'arch': 'x86_64',
         'epoch': None,
         'id': 12345678,
         'name': 'gnome-desktop3',
         'release': '1.module_2256+753d1eef',
         'version': '3.30.2'},
    ])

    p1 = patch('koji.read_config',
               return_value={'server':
                             'https://koji.fedoraproject.org/kojihub'})
    p2 = patch('koji.grab_session_options')
    p3 = patch('koji.ClientSession', return_value=session)

    p1.start()
    p2.start()
    p3.start()

    yield

    p1.stop()
    p2.stop()
    p3.stop()


def _generate_flatpak(rpm, flathub=None, common=False, expected_error_output=None):
    cmd = ['rpm2flatpak']
    cmd.append(rpm)
    if flathub:
        cmd += ['--flathub', flathub]
    if common:
        cmd += ['--flatpak-common']

    prevdir = os.getcwd()
    with tempfile.TemporaryDirectory() as workdir:
        try:
            os.chdir(workdir)
            runner = CliRunner()
            result = runner.invoke(_cli_commands, cmd, catch_exceptions=False)
            if expected_error_output is not None:
                assert result.exit_code != 0
                assert expected_error_output in result.output
                return
            else:
                assert result.exit_code == 0

            modulemd_fname = rpm + '.yaml'
            with open(rpm + '.yaml') as f:
                contents = f.read()

            log.info('%s:\n%s\n', modulemd_fname, contents)

            objs = Modulemd.objects_from_string(contents)
            assert len(objs) == 1
            assert isinstance(objs[0], Modulemd.Module)
            modmd = objs[0]

            with open('container.yaml') as f:
                contents = f.read()

            log.info('container.yaml:\n%s\n', contents)
            container_yaml = yaml.safe_load(contents)
        finally:
            os.chdir(prevdir)

    return modmd, container_yaml


class TestFlatpak(object):
    @pytest.mark.filterwarnings('ignore::DeprecationWarning:koji')
    @pytest.mark.filterwarnings('ignore::PendingDeprecationWarning:koji')
    @pytest.mark.needs_metadata
    def test_generated_flatpak_files(self):
        with mock_koji():
            modmd, container_yaml = _generate_flatpak('eog')

        # Expected description for 'eog'
        assert modmd.props.summary == "Eye of GNOME image viewer"
        assert modmd.props.description.startswith(
            "The Eye of GNOME image viewer (eog) is")

        # Expected licenses for 'eog'
        assert modmd.props.module_licenses.get() == ['MIT']
        assert modmd.props.content_licenses.get() == []

        # No API section for Flatpaks
        assert sorted(modmd.props.rpm_api.get()) == []

        # Expected components
        assert set(modmd.props.components_rpm) == set([
            'bubblewrap', 'eog', 'exempi', 'gnome-desktop3', 'libpeas'
        ])

        assert modmd.props.components_rpm['eog'].props.rationale == "Application package"
        assert modmd.props.components_rpm['eog'].props.buildorder == 10
        assert modmd.props.components_rpm['exempi'].props.rationale == "Runtime dependency"
        assert modmd.props.components_rpm['exempi'].props.buildorder == 0

        # Expected module dependencies for eog
        dependencies = modmd.props.dependencies
        assert len(dependencies) == 1

        buildrequires = dependencies[0].props.buildrequires
        assert set(buildrequires) == {'flatpak-runtime', 'platform'}
        assert buildrequires['flatpak-runtime'].get() == ['f29']
        assert buildrequires['platform'].get() == ['f29']

        requires = dependencies[0].props.requires
        assert set(requires) == {'flatpak-runtime', 'platform'}
        assert requires['flatpak-runtime'].get() == ['f29']
        assert requires['platform'].get() == ['f29']

    @responses.activate
    @pytest.mark.needs_metadata
    @pytest.mark.parametrize(('search_term', 'extension', 'expected_error'),
                             [
                                 ('org.gnome.eog', 'yaml', None),
                                 ('org.gnome.eog', 'yml', None),
                                 ('org.gnome.eog', 'json', None),
                                 ('eYe of gNome', 'yaml', None),
                                 ('org.gnome', 'yaml',
                                  'Multiple matches found on flathub.org'),
                                 ('notexist', 'yaml',
                                  'No match found on flathub.org'),
                             ])
    def test_flatpak_from_flathub(self, search_term, extension,
                                  expected_error):
        responses.add(responses.GET, 'https://flathub.org/api/v1/apps',
                      body=APPS_JSON, content_type='application/json')

        app_id = 'org.gnome.eog'
        base = 'https://raw.githubusercontent.com/flathub'

        for ext, content_type, body in [
                ('yml', 'application/x-yaml', EOG_YAML),
                ('yaml', 'application/x-yaml', EOG_YAML),
                ('json', 'application/json', EOG_JSON)]:
            if extension == ext:
                responses.add(responses.GET,
                              f"{base}/{app_id}/master/{app_id}.{ext}",
                              body=body, content_type=content_type)
            else:
                responses.add(responses.GET,
                              f"{base}/{app_id}/master/{app_id}.{ext}",
                              body='Not found', status=404)

        with mock_koji():
            if expected_error is None:
                modmd, container_yaml = \
                    _generate_flatpak('eog',
                                      flathub=search_term,
                                      expected_error_output=expected_error)

                f = container_yaml['flatpak']

                assert f['id'] == 'org.gnome.eog'
                assert f['command'] == 'eog'
                assert f['rename-desktop-file'] == 'eog.desktop'
                assert f['rename-appdata-file'] == 'eog.appdata.xml'
                assert f['rename-icon'] == 'eog'
                assert f['copy-icon'] is True
                assert f['finish-args'] == '--share=ipc\n--socket=x11'
            else:
                _generate_flatpak('eog', flathub=search_term,
                                  expected_error_output=expected_error)

    def test_flatpak_with_common(self):
        with mock_koji():
            modmd, container_yaml = _generate_flatpak('eog', common=True)

        # Expected components - does not include gnome-desktop3, since that's
        # in the mocked flatpak-common module
        assert set(modmd.props.components_rpm) == set([
            'bubblewrap', 'eog', 'exempi', 'libpeas'
        ])

        # Expected module dependencies for eog
        dependencies = modmd.props.dependencies
        assert len(dependencies) == 1

        buildrequires = dependencies[0].props.buildrequires
        assert set(buildrequires) == {'flatpak-runtime', 'flatpak-common', 'platform'}
        assert buildrequires['flatpak-common'].get() == ['f29']
        assert buildrequires['flatpak-runtime'].get() == ['f29']
        assert buildrequires['platform'].get() == ['f29']

        requires = dependencies[0].props.requires
        assert set(requires) == {'flatpak-runtime', 'flatpak-common', 'platform'}
        assert requires['flatpak-common'].get() == ['f29']
        assert requires['flatpak-runtime'].get() == ['f29']
        assert requires['platform'].get() == ['f29']
